/*
 * Project 64 - A Nintendo 64 emulator.
 *
 * (c) Copyright 2001 zilmar (zilmar@emulation64.com) and 
 * Jabo (jabo@emulation64.com).
 *
 * pj64 homepage: www.pj64.net
 *
 * Permission to use, copy, modify and distribute Project64 in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Project64 is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Project64 or software derived from Project64.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so if they want them.
 *
 */
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include "main.h"
#include "cpu.h"
#include "memory.h"
#include "plugin.h"
#include "psftag.h"

#define MenuLocOfUsedFiles	9
#define MenuLocOfUsedDirs	(MenuLocOfUsedFiles + 1)

#define SaveStateSize (4204380*2)

extern DWORD deflen, deffade, DefaultLength;
extern double defrelvol;
DWORD RomFileSize, RomRamSize, RomSaveUsing, RomCPUType, RomSelfMod, 
	RomUseTlb, RomUseLinking, RomCF, RomUseLargeBuffer, RomUseCache,
	RomReadRomDirect;
char CurrentFileName[MAX_PATH+1], RomName[MAX_PATH+1], RomHeader[64];
char LastRoms[10][MAX_PATH+1], LastDirs[10][MAX_PATH+1];

BOOL IsValidRomImage ( BYTE Test[4] );
int ReadSparse( HANDLE hFile, BYTE * buffer );
int EnableCompare=0;

BYTE * savestatespace;

BOOL forceinterpreter;

void ByteSwapRom (void) {
	DWORD count;

	//SendMessage( hStatusWnd, SB_SETTEXT, 0, (LPARAM)"Byte swapping image" );
	switch (*((DWORD *)&ROM[0])) {
	case 0x12408037:
		for( count = 0 ; count < RomFileSize; count += 4 ) {
			ROM[count] ^= ROM[count+2];
			ROM[count + 2] ^= ROM[count];
			ROM[count] ^= ROM[count+2];			
			ROM[count + 1] ^= ROM[count + 3];
			ROM[count + 3] ^= ROM[count + 1];
			ROM[count + 1] ^= ROM[count + 3];			
		}
		break;
	case 0x40123780:
		for( count = 0 ; count < RomFileSize; count += 4 ) {
			ROM[count] ^= ROM[count+3];
			ROM[count + 3] ^= ROM[count];
			ROM[count] ^= ROM[count+3];			
			ROM[count + 1] ^= ROM[count + 2];
			ROM[count + 2] ^= ROM[count + 1];
			ROM[count + 1] ^= ROM[count + 2];			
		}
		break;
	case 0x80371240: break;
	default:
		DisplayError("ByteSwapRom: %X",ROM[0]);
	}
}

BOOL IsValidRomImage ( BYTE Test[4] ) {
	if ( *((DWORD *)&Test[0]) == 0x40123780 ) { return TRUE; }
	if ( *((DWORD *)&Test[0]) == 0x12408037 ) { return TRUE; }
	if ( *((DWORD *)&Test[0]) == 0x80371240 ) { return TRUE; }
	return FALSE;
}

BOOL IsValidUSF ( BYTE Test[4] ) {
	if ( *((DWORD *)&Test[0]) == 0x21465350 ) { return TRUE; } // "PSF!"
	return FALSE;
}

BOOL IsValidSparseRom ( BYTE Test[4] ) {
	if ( *((DWORD *)&Test[0]) == 0x34365253 ) { return TRUE; }
	return FALSE;
}

BOOL IsTagPresent ( BYTE Test[5] ) {
	if ( *((DWORD *)&Test[0]) == 0x4741545b && Test[4]==0x5D) {return TRUE;}
	return FALSE;
}

void LoadRomOptions ( void ) {
//	DWORD NewRamSize;

	RomRamSize        = 0x800000; //0x400000;
	RomSaveUsing      = Auto;
	RomCF             = -1;
	RomCPUType        = ((UseRecompiler && !forceinterpreter)?CPU_Recompiler:CPU_Interpreter); //CPU_Default;
	RomSelfMod        = ModCode_CheckMemoryCache; //ModCode_Default;
	RomUseTlb         = TRUE;
	RomUseCache       = TRUE;
	RomUseLargeBuffer = FALSE;
	RomReadRomDirect  = FALSE;
	RomUseLinking     = -1;

	RdramSize=RomRamSize;

	CPU_Type = RomCPUType;
	//if (CPU_Type != CPU_Interpreter) DisplayError("Using recompiler");
	CountPerOp = Default_CountPerOp;
	
	SelfModCheck = ModCode_Default;
	UseTlb = RomUseTlb;
	UseLinking = TRUE;
	DisableRegCaching = FALSE;
	Timer_Initialize((double)60);
}

int ReadUSF( const char * FileName ) {
	DWORD dwRead;
	//DWORD len;
	//DWORD start;
	DWORD reservestart;
	DWORD tagstart;
	DWORD taglen;
	DWORD codesize;
	DWORD reservesize;
	HANDLE hFile;
	BYTE length[256],fade[256],buf[256],rvol[256];
	char lib[_MAX_PATH],drive[_MAX_DRIVE] ,usfFileName[_MAX_DIR],dir[_MAX_DIR], ext[_MAX_EXT];
	char libpath[_MAX_PATH];
	
	BYTE Test[5];
	char *psftag;
	//char psftag[50001];

	hFile = CreateFile(FileName,GENERIC_READ,FILE_SHARE_READ,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
		NULL);

	if (hFile == INVALID_HANDLE_VALUE) { 
		DisplayError("Failed to open file\n%s",FileName);
		return 1;
	}

	ReadFile(hFile,Test,4,&dwRead,NULL);
	if (!IsValidUSF(Test)) {
		DisplayError("Not a USF file\n%s",FileName);
		CloseHandle(hFile);
		return 1;
	}
	// Read USF contents
	ReadFile(hFile,&reservesize,4,&dwRead,NULL); // size of reserved area
	ReadFile(hFile,&codesize,4,&dwRead,NULL); // size of program
	ReadFile(hFile,Test,4,&dwRead,NULL); // CRC (discarded)
	
	reservestart=SetFilePointer(hFile,0,0,FILE_CURRENT);
	tagstart=reservestart+codesize+reservesize;
	SetFilePointer(hFile,tagstart,0,FILE_BEGIN);

	ReadFile(hFile,Test,5,&dwRead,NULL);
	// read tags if there are any
	if (IsTagPresent(Test)) {
		tagstart+=5;
		taglen=SetFilePointer(hFile,0,0,FILE_END)-tagstart;
		SetFilePointer(hFile,tagstart,0,FILE_BEGIN);
		if (taglen > 0) {
			//psftag=malloc(taglen+2);
			psftag=malloc(50001);
					
			ReadFile(hFile,psftag,taglen,&dwRead,NULL);
			psftag[taglen]='\0';

			psftag_raw_getvar(psftag,"_lib",lib,sizeof(lib)-1); // no _libn support
			
			// revised so tags in usflib don't reset length
			psftag_raw_getvar(psftag,"length",length,sizeof(length)-1);
			if (!IgnoreTrackLength && strlen(length)) TrackLength=LengthFromString(length);
			if (TrackLength <= 0) {
				if (IgnoreTrackLength) TrackLength=0;
			}
			psftag_raw_getvar(psftag,"fade",fade,sizeof(fade)-1);
			if (!IgnoreTrackLength && strlen(fade)) {
				FadeLength=LengthFromString(fade);
				TrackLength+=FadeLength; // comply with PSF standard timing:
			}							 // "length" tag specifies length before fade,
										 // "fade" specifies length of fade

			psftag_raw_getvar(psftag,"volume",rvol,sizeof(rvol)-1);
			if (strlen(rvol)) {
				sscanf(rvol,"%lf",&relvol);
				relvol*=defrelvol;
			}
					
			// Check for bug tags:
			// _64thbug1: Pilot Wings64 can't be run under the recompiler as of 9/15/04
			// _64thbug2: Goldeneye can't be run under the recompiler as of 9/15/04
					
			//psftag_raw_getvar(psftag,"_64thbug1",buf,sizeof(buf)-1);
			//if (strlen(buf)>0) forceinterpreter=1;
			//psftag_raw_getvar(psftag,"_64thbug2",buf,sizeof(buf)-1);
			//if (strlen(buf)>0) forceinterpreter=1;
			
			// Most often you don't want the count=compare interrupt, but sometimes... (i.e. PD,BK!)
			psftag_raw_getvar(psftag,"_enablecompare",buf,sizeof(buf)-1);
			if (strlen(buf)>0) EnableCompare=1;
			
			free(psftag);
			psftag=NULL;
									
			if (strlen(lib) > 0) {
				_splitpath( FileName, drive, dir, usfFileName, ext );
				
				sprintf(libpath,"%s%s%s",drive,dir,lib);

				if (ReadUSF(libpath)) return 1;
			}
		}
	}

	if (DefaultLength && TrackLength <= 0) {
		TrackLength=(deflen+deffade)*1000;
		FadeLength=deffade*1000;
	}

	
	SetFilePointer(hFile,reservestart,0,FILE_BEGIN);

	if (codesize > 0) DisplayError("What's in code ?");
	if (reservesize > 0) {
		if (ReadSparse(hFile,ROM)) return 1;
		if (ReadSparse(hFile,savestatespace)) return 1;
	}		

	CloseHandle(hFile);
	return 0;
}

int ReadSparse( HANDLE hFile, BYTE * buffer ) {
	DWORD dwRead;
	//DWORD FileSize;
	DWORD len;
	DWORD start;
	//DWORD codestart;
	//DWORD tagstart;
	//DWORD taglen;
	BYTE Test[5];
	//char *psftag;

	ReadFile(hFile,Test,4,&dwRead,NULL); // SR64 header
	if (!IsValidSparseRom(Test)) {
		if (*((unsigned long*)Test)==0) return 0; // empty section
		DisplayError("Bad sparse ROM");
		CloseHandle(hFile);
		return 1;
	}

	ReadFile(hFile,&len,4,&dwRead,NULL); // first len
	while (len) {
		ReadFile(hFile,&start,4,&dwRead,NULL); // start addr
		//DisplayError("len=%08x start=%08x",len,start);
		if (!ReadFile(hFile,&buffer[start],len,&dwRead,NULL)) {
			DisplayError("Failed to copy sparse to memory\nlen=%08x",len);
			CloseHandle( hFile );
			return 1;
		}
		ReadFile(hFile,&len,4,&dwRead,NULL);
	}
	return 0;
}

int OpenChosenFile ( void ) {
	//#define ReadFromRomSection	0x400000
	DWORD count;
	forceinterpreter=0;
	EnableCompare=0;
	//HANDLE hSaveFile;
	//int count;
	
	CloseCpu();

	RomFileSize = 0x2000000; //0x800000;
		
	if (!Allocate_ROM()) {
		DisplayError("Not enough memory to load ROM");
		return 1;
	}

	savestatespace = VirtualAlloc(NULL,SaveStateSize,MEM_RESERVE|MEM_COMMIT,PAGE_READWRITE);
	if (!savestatespace) {
		DisplayError("Not enough memory to load save state");
	  return 1;
	}


	TrackLength=0;
	FadeLength=0;
	relvol=defrelvol;
	if (ReadUSF( CurrentFileName )) return 1;

	memcpy(RomHeader,ROM,sizeof(RomHeader));
		
	memcpy(&RomName[0],(void *)(ROM + 0x20),20);

	CloseCpu();

	memset(&CPU_Action,0,sizeof(CPU_Action));
	CPU_Action.hStepping = CreateEvent( NULL, FALSE, FALSE, NULL);
	WrittenToRom = FALSE;

	InitilizeTLB();
	//InitalizeR4300iRegisters(LoadPifRom(*(ROM + 0x3D)),*(ROM + 0x3D),GetCicChipID(ROM));
	SetFpuLocations();

	RecompPos = RecompCode;

	Timers.CurrentTimerType = -1;
	Timers.Timer = 0;
	CurrentFrame = 0;
	CurrentPercent = 0;
	for (count = 0; count < MaxTimers; count ++) { Timers.Active[count] = FALSE; }
	ChangeTimer(ViTimer,5000); 
	ViFieldNumber = 0;
	DMAUsed = FALSE;
	CPU_Paused = FALSE;
	ManualPaused = FALSE;
	Timer_Start();
	LoadRomOptions();

	if (CPU_Type == CPU_Interpreter) BuildInterpreter();
	else {
		if (!AllocateRecompMemory()) return 1;
	}

	LoadSaveState();
	

	VirtualFree(savestatespace,0,MEM_RELEASE);

	StartEmulation();
	return 0;
}

void ReadFromMem(const void * source, void * target, DWORD length, DWORD *offset) {
	memcpy((BYTE*)target,((BYTE*)source)+*offset,length);
	*offset+=length;
}

void LoadSaveState() {
	DWORD Value,SaveRDRAMSize,offset;
	BYTE LoadHeader[0x40];
	int count;

	offset=0;
	ReadFromMem(savestatespace,&Value,sizeof(Value),&offset);
	if (Value != 0x23D8A6C8) { DisplayError("Improper save state"); return; }

	ReadFromMem( savestatespace,&SaveRDRAMSize,sizeof(SaveRDRAMSize),&offset);
	ReadFromMem( savestatespace,LoadHeader,0x40,&offset);

	if (CPU_Type == CPU_Recompiler) ResetRecompCode(); 
	
	Timers.CurrentTimerType = -1;
	Timers.Timer = 0;
	for (count = 0; count < MaxTimers; count ++) { Timers.Active[count] = FALSE; }

	//fix rdram size
	//RdramSize = //SaveRDRAMSize;

	ReadFromMem( savestatespace,&Value,sizeof(Value),&offset);
	ChangeTimer(ViTimer,Value);
	ReadFromMem( savestatespace,&PROGRAM_COUNTER,sizeof(PROGRAM_COUNTER),&offset);
	ReadFromMem( savestatespace,GPR,sizeof(_int64)*32,&offset);
	ReadFromMem( savestatespace,FPR,sizeof(_int64)*32,&offset);
	ReadFromMem( savestatespace,CP0,sizeof(DWORD)*32,&offset);
	ReadFromMem( savestatespace,FPCR,sizeof(DWORD)*32,&offset);
	ReadFromMem( savestatespace,&HI,sizeof(_int64),&offset);
	ReadFromMem( savestatespace,&LO,sizeof(_int64),&offset);
	ReadFromMem( savestatespace,RegRDRAM,sizeof(DWORD)*10,&offset);
	ReadFromMem( savestatespace,RegSP,sizeof(DWORD)*10,&offset);
	ReadFromMem( savestatespace,RegDPC,sizeof(DWORD)*10,&offset);
	ReadFromMem( savestatespace,RegMI,sizeof(DWORD)*4,&offset);
	ReadFromMem( savestatespace,RegVI,sizeof(DWORD)*14,&offset);
	ReadFromMem( savestatespace,RegAI,sizeof(DWORD)*6,&offset);
	ReadFromMem( savestatespace,RegPI,sizeof(DWORD)*13,&offset);
	ReadFromMem( savestatespace,RegRI,sizeof(DWORD)*8,&offset);
	ReadFromMem( savestatespace,RegSI,sizeof(DWORD)*4,&offset);
	ReadFromMem( savestatespace,tlb,sizeof(TLB)*32,&offset);
	ReadFromMem( savestatespace,PIF_Ram,0x40,&offset);
	//ReadFromMem( savestatespace,RDRAM,RdramSize,&offset);
	//DisplayError("SaveRDRAMSize=%#08x",SaveRDRAMSize);
	ReadFromMem( savestatespace,RDRAM,SaveRDRAMSize,&offset);
	ReadFromMem( savestatespace,DMEM,0x1000,&offset);
	ReadFromMem( savestatespace,IMEM,0x1000,&offset);

	ChangeCompareTimer();
	if (AiRomClosed != NULL)   { AiRomClosed(); }
	if (RSPRomClosed != NULL) { RSPRomClosed(); }
	DlistCount = 0;
	AlistCount = 0;
	AI_STATUS_REG = 0;
	AiDacrateChanged(SYSTEM_NTSC);
	SetupTLB();
	CheckInterrupts();
	DMAUsed = TRUE;
	return;
}