/*
 * Project 64 - A Nintendo 64 emulator.
 *
 * (c) Copyright 2001 zilmar (zilmar@emulation64.com) and 
 * Jabo (jabo@emulation64.com).
 *
 * pj64 homepage: www.pj64.net
 *
 * Permission to use, copy, modify and distribute Project64 in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Project64 is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Project64 or software derived from Project64.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so if they want them.
 *
 */
#define DefaultRSPDll				"RSP.dll"
//#define DefaultAudioDll				"AziSnd.dll"

#define PLUGIN_TYPE_RSP				1
#define PLUGIN_TYPE_GFX				2
#define PLUGIN_TYPE_AUDIO			3
#define PLUGIN_TYPE_CONTROLLER		4

#define SYSTEM_NTSC					0
#define SYSTEM_PAL					1
#define SYSTEM_MPAL					2

#define EXPORT
#define CALL						_cdecl

typedef struct {
	WORD Version;        /* Should be set to 1 */
	WORD Type;           /* Set to PLUGIN_TYPE_GFX */
	char Name[100];      /* Name of the DLL */

	/* If DLL supports memory these memory options then set them to TRUE or FALSE
	   if it does not support it */
	BOOL NormalMemory;   /* a normal BYTE array */ 
	BOOL MemoryBswaped;  /* a normal BYTE array where the memory has been pre
	                          bswap on a dword (32 bits) boundry */
} PLUGIN_INFO;

typedef struct {
	HINSTANCE hInst;
	BOOL MemoryBswaped;    /* If this is set to TRUE, then the memory has been pre
	                          bswap on a dword (32 bits) boundry */
	BYTE * RDRAM;
	BYTE * DMEM;
	BYTE * IMEM;

	DWORD * MI__INTR_REG;

	DWORD * SP__MEM_ADDR_REG;
	DWORD * SP__DRAM_ADDR_REG;
	DWORD * SP__RD_LEN_REG;
	DWORD * SP__WR_LEN_REG;
	DWORD * SP__STATUS_REG;
	DWORD * SP__DMA_FULL_REG;
	DWORD * SP__DMA_BUSY_REG;
	DWORD * SP__PC_REG;
	DWORD * SP__SEMAPHORE_REG;

	DWORD * DPC__START_REG;
	DWORD * DPC__END_REG;
	DWORD * DPC__CURRENT_REG;
	DWORD * DPC__STATUS_REG;
	DWORD * DPC__CLOCK_REG;
	DWORD * DPC__BUFBUSY_REG;
	DWORD * DPC__PIPEBUSY_REG;
	DWORD * DPC__TMEM_REG;

	void ( __cdecl *CheckInterrupts)( void );
	void (__cdecl *ProcessDlist)( void );
	void (__cdecl *ProcessAlist)( void );
	void (__cdecl *ProcessRdpList)( void );

} RSP_INFO_1_0;

typedef struct {
	BOOL Present;
	BOOL RawData;
	int  Plugin;
} CONTROL;

typedef struct {
	HWND hMainWindow;
	HINSTANCE hinst;

	BOOL MemoryBswaped;		// If this is set to TRUE, then the memory has been pre
							//   bswap on a dword (32 bits) boundry, only effects header. 
							//	eg. the first 8 bytes are stored like this:
							//        4 3 2 1   8 7 6 5
	BYTE * HEADER;			// This is the rom header (first 40h bytes of the rom)
	CONTROL *Controls;		// A pointer to an array of 4 controllers .. eg:
							// CONTROL Controls[4];
} CONTROL_INFO;

typedef struct {
	HINSTANCE hInst;
	BOOL MemoryBswaped;    /* If this is set to TRUE, then the memory has been pre
	                          bswap on a dword (32 bits) boundry */
	BYTE * RDRAM;
	BYTE * DMEM;
	BYTE * IMEM;

	DWORD * MI__INTR_REG;

	DWORD * SP__MEM_ADDR_REG;
	DWORD * SP__DRAM_ADDR_REG;
	DWORD * SP__RD_LEN_REG;
	DWORD * SP__WR_LEN_REG;
	DWORD * SP__STATUS_REG;
	DWORD * SP__DMA_FULL_REG;
	DWORD * SP__DMA_BUSY_REG;
	DWORD * SP__PC_REG;
	DWORD * SP__SEMAPHORE_REG;

	DWORD * DPC__START_REG;
	DWORD * DPC__END_REG;
	DWORD * DPC__CURRENT_REG;
	DWORD * DPC__STATUS_REG;
	DWORD * DPC__CLOCK_REG;
	DWORD * DPC__BUFBUSY_REG;
	DWORD * DPC__PIPEBUSY_REG;
	DWORD * DPC__TMEM_REG;

	void ( __cdecl *CheckInterrupts)( void );
	void (__cdecl *ProcessDlist)( void );
	void (__cdecl *ProcessAlist)( void );
	void (__cdecl *ProcessRdpList)( void );
	void (__cdecl *ShowCFB)( void );
	void (__cdecl *IsSeeking)( void ); 
} RSP_INFO_1_1;

typedef struct {
	void (__cdecl *UpdateBreakPoints)( void );
	void (__cdecl *UpdateMemory)( void );
	void (__cdecl *UpdateR4300iRegisters)( void );
	void (__cdecl *Enter_BPoint_Window)( void );
	void (__cdecl *Enter_R4300i_Commands_Window)( void );
	void (__cdecl *Enter_R4300i_Register_Window)( void );
	void (__cdecl *Enter_RSP_Commands_Window) ( void );
	void (__cdecl *Enter_Memory_Window)( void );
} DEBUG_INFO;

typedef struct {
	HWND hwnd;
	HINSTANCE hinst;

	BOOL MemoryBswaped;    // If this is set to TRUE, then the memory has been pre
	                       //   bswap on a dword (32 bits) boundry 
						   //	eg. the first 8 bytes are stored like this:
	                       //        4 3 2 1   8 7 6 5
	BYTE * HEADER;	// This is the rom header (first 40h bytes of the rom
					// This will be in the same memory format as the rest of the memory.
	BYTE * RDRAM;
	BYTE * DMEM;
	BYTE * IMEM;

	DWORD * MI__INTR_REG;

	DWORD * AI__DRAM_ADDR_REG;
	DWORD * AI__LEN_REG;
	DWORD * AI__CONTROL_REG;
	DWORD * AI__STATUS_REG;
	DWORD * AI__DACRATE_REG;
	DWORD * AI__BITRATE_REG;

	In_Module * mod;

	void (__cdecl *CheckInterrupts)( void );
} AUDIO_INFO;

/*** Conteroller plugins ****/
#define PLUGIN_NONE					1
#define PLUGIN_MEMPAK				2
#define PLUGIN_RUMBLE_PAK			3 
#define PLUGIN_TANSFER_PAK			4 // not implemeted for non raw data
#define PLUGIN_RAW					5 // the controller plugin is passed in raw data

/******** All DLLs have this function **************/
void (__cdecl *GetDllInfo)             ( PLUGIN_INFO * PluginInfo );

/********** RSP DLL: Functions *********************/
//void (__cdecl *GetRspDebugInfo)    ( RSPDEBUG_INFO * DebugInfo );
void (__cdecl *RSPCloseDLL)        ( void );
void (__cdecl *RSPDllAbout)        ( HWND hWnd );
void (__cdecl *RSPDllConfig)       ( HWND hWnd );
void (__cdecl *RSPRomClosed)       ( void );
DWORD (__cdecl *DoRspCycles)       ( DWORD );
void (__cdecl *InitiateRSP_1_0)    ( RSP_INFO_1_0 Rsp_Info, DWORD * Cycles);
void (__cdecl *InitiateRSP_1_1)    ( RSP_INFO_1_1 Rsp_Info, DWORD * Cycles);
//void (__cdecl *InitiateRSPDebugger)( DEBUG_INFO DebugInfo);

/********** RSP LIB: Functions ***********/
void __cdecl RspCloseDLL (void);
void __cdecl RspDllAbout ( HWND );
DWORD __cdecl RspDoRspCycles ( DWORD  );
void __cdecl RspGetDllInfo ( PLUGIN_INFO * );
void __cdecl RspInitiateRSP ( RSP_INFO_1_1 , DWORD * );
void __cdecl RspRomClosed (void);
void __cdecl RspDllConfig (HWND);
void __cdecl RspAudioHLE ( DWORD );

/************ Audio DLL: Functions *****************/
//void (__cdecl *AiCloseDLL)       ( void );
//void (__cdecl *AiDacrateChanged) ( int SystemType );
//void (__cdecl *AiLenChanged)     ( void );
//void (__cdecl *AiDllAbout)       ( HWND hParent );
//void (__cdecl *AiDllConfig)      ( HWND hParent );
//void (__cdecl *AiDllTest)        ( HWND hParent );
//DWORD (__cdecl *AiReadLength)    ( void );
//void (__cdecl *AiRomClosed)      ( void );
//void (__cdecl *AiUpdate)         ( BOOL Wait );
//BOOL (__cdecl *InitiateAudio)    ( AUDIO_INFO Audio_Info );
//void (__cdecl *ProcessAList)     ( void );

/************ Audio LIB: Functions **********/
EXPORT void CALL AiDacrateChanged (int  SystemType);
EXPORT void CALL AiLenChanged (void);
EXPORT DWORD CALL AiReadLength (void);
//EXPORT void CALL AiUpdate (BOOL Wait);
EXPORT void CALL AiCloseDLL (void);
//EXPORT void CALL AziDllAbout ( HWND hParent );
//EXPORT void CALL AziDllConfig ( HWND hParent );
//EXPORT void CALL AziDllTest ( HWND hParent );
EXPORT void CALL AiGetDllInfo ( PLUGIN_INFO * PluginInfo );
EXPORT BOOL CALL InitiateAudio (AUDIO_INFO Audio_Info);
EXPORT void CALL ProcessAList(void);
EXPORT void CALL AiRomClosed (void);
//EXPORT void CALL AziAiCallBack (void);
EXPORT DWORD CALL IsFreqSet(void);
EXPORT DWORD CALL IsSeeking(void);
EXPORT DWORD CALL PlayTime(void);
EXPORT void CALL SetVolume(int newvol);
EXPORT void CALL SeekTo(int seek_dest);

/********** Plugin: Functions *********************/
void GetPluginDir        ( char * Directory );
void PluginConfiguration ( HWND hWnd );
void SetupPlugins        ( void );
//void SetupPluginScreen   ( void );
void ShutdownPlugins     ( void );

/********** External Global Variables ***************/
#define MaxDlls	100
extern char RspDLL[100], AudioDLL[100], * PluginNames[MaxDlls];
extern DWORD PluginCount, RspTaskValue, AudioIntrReg;
extern CONTROL Controllers[4];
extern WORD RSPVersion;
