/*
 * Project 64 - A Nintendo 64 emulator.
 *
 * (c) Copyright 2001 zilmar (zilmar@emulation64.com) and 
 * Jabo (jabo@emulation64.com).
 *
 * pj64 homepage: www.pj64.net
 *
 * Permission to use, copy, modify and distribute Project64 in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Project64 is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Project64 or software derived from Project64.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so if they want them.
 *
 */

#include <Windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <shlobj.h>
#include <stdio.h>
#include "main.h"
#include "cpu.h"
#include "plugin.h"
#include "psftag.h"
#include "resource.h"
#include <mmreg.h>
#include <msacm.h>
#include <math.h>

// post this to the main window at end of file (after playback as stopped)
#define WM_WA_MPEG_EOF WM_USER+2

#define ProgName "64th Note v0.05"
#define AppVer   ProgName "\nBased on Project 64 by Zilmar and Jabo\nAudio by Azimer\nPSF concept and tagging by Neill Corlett\nPlayer by hcs"
#define AppName  ProgName " (x86)"
#define AppReg   "Software\\Halley's Comet Software\\" ProgName

//LARGE_INTEGER Frequency, Frames[NoOfFrames], LastFrame;
BOOL HaveDebugger, AutoLoadMapFile, ShowUnhandledMemory, ShowTLBMisses=1, 
	ShowDListAListCount, ShowCompMem, Profiling, IndvidualBlock, AutoStart, 
	AutoSleep, DisableRegCaching, UseIni, UseTlb, UseLinking, RomBrowser,
	IgnoreMove, Rercursion, ShowPifRamErrors, LimitFPS, ShowCPUPer, AutoZip, 
	AutoFullScreen, SystemABL;
DWORD CurrentFrame, CPU_Type, SystemCPU_Type, SelfModCheck, SystemSelfModCheck, 
	RomsToRemember, RomDirsToRemember;
//HWND hMainWindow, hHiddenWin, hStatusWnd;
char CurrentSave[256];
//HMENU hMainMenu;
HINSTANCE hInst;
DWORD WaitMode=0;
DWORD TrackLength=0;
DWORD FadeLength=0;
DWORD AudioHLE=0, FastSeek=0, IgnoreTrackLength=0, UseRecompiler=1, DetectSilence=1, DefaultLength=1;
DWORD deflen=180,deffade=4,silencelength=5,silencedetected=0;
char titlefmt[256]="%game% - %title%";

char *infoDlgpsftag=NULL;
//char infoDlgpsftag[50001];
char *getfileinfopsftag=NULL;
//char getfileinfopsftag[50001];
int infoDlgWriteNewTags=0;

extern In_Module mod;
char lastfn[MAX_PATH];
//int killDecodeThread=0;
int paused=0;

HANDLE thread_handle=INVALID_HANDLE_VALUE;

void __cdecl DisplayError (char * Message, ...) {
	char Msg[400];
	va_list ap;

	va_start( ap, Message );
	vsprintf( Msg, Message, ap );
	va_end( ap );
	MessageBox(NULL,Msg,"Error",MB_OK|MB_ICONERROR|MB_SETFOREGROUND);
	//SetActiveWindow(hMainWindow);
}

int InitalizeApplication ( HINSTANCE hInstance ) {
	HKEY hKeyResults = 0;
	//char String[200];
	//long lResult;

	//IntComStruct.dwSize = sizeof(IntComStruct);
	//IntComStruct.dwICC = ICC_TREEVIEW_CLASSES;
	//InitCommonControls();	
	//InitCommonControlsEx(&IntComStruct);
	hInst = hInstance;

	//DisplayError("Allocate_Memory succeeded");

	//hPauseMutex = CreateMutex(NULL,FALSE,NULL);
	
	IgnoreMove = FALSE;
	//CPU_Type = Default_CPU;
	CPU_Type = SystemCPU_Type = (UseRecompiler?CPU_Recompiler:CPU_Interpreter); //Default_CPU;
	SystemSelfModCheck = Default_SelfModCheck;
	SystemRdramSize = Default_RdramSize;
	SystemABL = Default_AdvancedBlockLink;
	AutoStart = Default_AutoStart;
	AutoSleep = Default_AutoSleep;
	DisableRegCaching = Default_DisableRegCaching;
	UseIni = Default_UseIni;
	AutoZip = Default_AutoZip;
	AutoFullScreen = FALSE;
	RomsToRemember = Default_RomsToRemember;
	RomDirsToRemember = Default_RomsDirsToRemember;
	AutoLoadMapFile = Default_AutoMap;
	ShowUnhandledMemory = Default_ShowUnhandledMemory;
	ShowCPUPer = Default_ShowCPUPer;
	LimitFPS = Default_LimitFPS;
	ShowTLBMisses = Default_ShowTLBMisses;
	Profiling = Default_ProfilingOn;
	IndvidualBlock = Default_IndvidualBlock;
	RomBrowser = Default_UseRB;
	Rercursion = Default_Rercursion;

	if (!Allocate_Memory()) { 
		DisplayError("Failed to allocate Memory");
		return FALSE; 
	}

	SetupRegisters(&Registers);
	//QueryPerformanceFrequency(&Frequency);
	return TRUE;
}

void ShutdownApplication ( void ) {
	CloseCpu();
	
	ShutdownPlugins();
	Release_Memory();
	//CloseHandle(hPauseMutex);
}

// Thanks to
// http://www4.ncsu.edu:8030/~jgbishop/codetips/general/registry_access_advanced.html
// for an excellent introduction to the registry.

void WriteSettings(void) {
	HKEY hKey;
	DWORD dwDisp;

	RegCreateKeyEx(HKEY_CURRENT_USER,AppReg,0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&dwDisp);
	RegSetValueEx(hKey,"Title Format",0,REG_SZ,titlefmt,strlen(titlefmt));
	RegSetValueEx(hKey,"Fast Seek",0,REG_DWORD,(BYTE*)&FastSeek,sizeof(FastSeek));
	RegSetValueEx(hKey,"Audio HLE",0,REG_DWORD,(BYTE*)&AudioHLE,sizeof(AudioHLE));
	RegSetValueEx(hKey,"Ignore Track Limits",0,REG_DWORD,(BYTE*)&IgnoreTrackLength,sizeof(IgnoreTrackLength));
	RegSetValueEx(hKey,"Default Length",0,REG_DWORD,(BYTE*)&DefaultLength,sizeof(DefaultLength));
	RegSetValueEx(hKey,"Default Length Value",0,REG_DWORD,(BYTE*)&deflen,sizeof(deflen));
	RegSetValueEx(hKey,"Default Fade Value",0,REG_DWORD,(BYTE*)&deffade,sizeof(deflen));
	RegSetValueEx(hKey,"Detect Silence",0,REG_DWORD,(BYTE*)&DetectSilence,sizeof(DetectSilence));
	RegSetValueEx(hKey,"Detect Silence Value",0,REG_DWORD,(BYTE*)&silencelength,sizeof(silencelength));
	RegSetValueEx(hKey,"Use Recompiler",0,REG_DWORD,(BYTE*)&UseRecompiler,sizeof(UseRecompiler));

	RegCloseKey(hKey);
}

void ReadSettings(void) {
	HKEY hKey;
	long lResult;
	unsigned long dwDataType,bufferlength;
	lResult=RegOpenKeyEx(HKEY_CURRENT_USER,AppReg,0,KEY_QUERY_VALUE,&hKey);
	if (lResult == ERROR_SUCCESS) {
		bufferlength = sizeof(titlefmt);
		RegQueryValueEx(hKey,"Title Format",0,&dwDataType,titlefmt,&bufferlength);
		bufferlength = sizeof(FastSeek);
		RegQueryValueEx(hKey,"Fast Seek",0,&dwDataType,(BYTE*)&FastSeek,&bufferlength);
		bufferlength = sizeof(AudioHLE);
		RegQueryValueEx(hKey,"Audio HLE",0,&dwDataType,(BYTE*)&AudioHLE,&bufferlength);
		bufferlength = sizeof(IgnoreTrackLength);
		RegQueryValueEx(hKey,"Ignore Track Limits",0,&dwDataType,(BYTE*)&IgnoreTrackLength,&bufferlength);
		
		bufferlength = sizeof(DefaultLength);
		RegQueryValueEx(hKey,"Default Length",0,&dwDataType,(BYTE*)&DefaultLength,&bufferlength);
		
		bufferlength = sizeof(deflen);
		RegQueryValueEx(hKey,"Default Length Value",0,&dwDataType,(BYTE*)&deflen,&bufferlength);

		bufferlength = sizeof(deffade);
		RegQueryValueEx(hKey,"Default Fade Value",0,&dwDataType,(BYTE*)&deffade,&bufferlength);
		
		bufferlength = sizeof(DetectSilence);
		RegQueryValueEx(hKey,"Detect Silence",0,&dwDataType,(BYTE*)&DetectSilence,&bufferlength);

		bufferlength = sizeof(silencelength);
		RegQueryValueEx(hKey,"Detect Silence Value",0,&dwDataType,(BYTE*)&silencelength,&bufferlength);

		bufferlength = sizeof(UseRecompiler);
		RegQueryValueEx(hKey,"Use Recompiler",0,&dwDataType,(BYTE*)&UseRecompiler,&bufferlength);

		RegCloseKey(hKey);
	} else {
		WriteSettings();
	}
}

BOOL CALLBACK configDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	char buf[256];
	switch (uMsg) {	
	case WM_CLOSE:
		EndDialog(hDlg,TRUE);
		return 0;
	case WM_INITDIALOG:
		if (FastSeek) CheckDlgButton(hDlg,IDC_FASTSEEK,BST_CHECKED);
		if (AudioHLE) CheckDlgButton(hDlg,IDC_AUDIOHLE,BST_CHECKED);
		if (IgnoreTrackLength) CheckDlgButton(hDlg,IDC_NOLENGTH,BST_CHECKED);
		if (DefaultLength) CheckDlgButton(hDlg,IDC_DEFLEN,BST_CHECKED);
		if (DetectSilence) CheckDlgButton(hDlg,IDC_DETSIL,BST_CHECKED);
		if (UseRecompiler) CheckDlgButton(hDlg,IDC_RECOMPILER,BST_CHECKED);
		SetDlgItemText(hDlg,IDC_TITLEFMT,titlefmt);
		sprintf(buf,"%i",deflen);
		SetDlgItemText(hDlg,IDC_DEFLENVAL,buf);
		sprintf(buf,"%i",deffade);
		SetDlgItemText(hDlg,IDC_DEFFADEVAL,buf);
		sprintf(buf,"%i",silencelength);
		SetDlgItemText(hDlg,IDC_DETSILVAL,buf);
		break;
	case WM_COMMAND:
		switch (GET_WM_COMMAND_ID(wParam, lParam)) {
		case IDOK:
			FastSeek=(IsDlgButtonChecked(hDlg, IDC_FASTSEEK) == BST_CHECKED) ? TRUE : FALSE;
			AudioHLE=(IsDlgButtonChecked(hDlg, IDC_AUDIOHLE) == BST_CHECKED) ? TRUE : FALSE;
			IgnoreTrackLength=(IsDlgButtonChecked(hDlg, IDC_NOLENGTH) == BST_CHECKED) ? TRUE : FALSE;
			DefaultLength=(IsDlgButtonChecked(hDlg, IDC_DEFLEN) == BST_CHECKED) ? TRUE : FALSE;
			UseRecompiler=(IsDlgButtonChecked(hDlg, IDC_RECOMPILER) == BST_CHECKED) ? TRUE : FALSE;
			DetectSilence=(IsDlgButtonChecked(hDlg, IDC_DETSIL) == BST_CHECKED) ? TRUE : FALSE;
			RspAudioHLE(AudioHLE);
			GetDlgItemText(hDlg,IDC_TITLEFMT,titlefmt,sizeof(titlefmt)-1);
			GetDlgItemText(hDlg,IDC_DEFLENVAL,buf,sizeof(buf)-1);
			sscanf(buf,"%i",&deflen);
			GetDlgItemText(hDlg,IDC_DEFFADEVAL,buf,sizeof(buf)-1);
			sscanf(buf,"%i",&deffade);
			GetDlgItemText(hDlg,IDC_DETSILVAL,buf,sizeof(buf)-1);
			sscanf(buf,"%i",&silencelength);
			WriteSettings();
		case IDCANCEL:
			EndDialog(hDlg,TRUE);
			break;
		}
	default:
		return 0;
	}

	return 1;
}

void config(HWND hwndParent)
{
	//MessageBox(hwndParent,"No config for you!","Configuration",MB_OK);
	DialogBox(mod.hDllInstance, (const char *)IDD_CONFIG_WINDOW, hwndParent, configDlgProc);
}
void about(HWND hwndParent)
{
	MessageBox(hwndParent,AppVer,"About",MB_OK);
}

void init() { ReadSettings(); }

void quit() { WriteSettings();}

int isourfile(char *fn) 
{ 
	return 0;
}

int play(char *fn) { 
	//int tmp;
	strcpy(lastfn,fn);

	paused=0;
	silencedetected=0;
	//memset(sample_buffer,0,sizeof(sample_buffer));

	//if (!killDecodeThread) {
	//	ShutdownApplication ();	
	//}
	//ShutdownApplication();

	if ( !InitalizeApplication (mod.hDllInstance) ) {
		ShutdownApplication();
		PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);
		return FALSE;
	}
	SetupPlugins();
	strcpy(CurrentFileName,lastfn);

	RspAudioHLE(AudioHLE);
	
	//killDecodeThread=0;
	//thread_handle = (HANDLE) CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) PlayThread,(void *) &killDecodeThread,0,&tmp);
	if (OpenChosenFile()) {
		ShutdownApplication();
		PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);
		return FALSE;
	}
	//CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)OpenChosenFile,NULL,0, &ThreadID);	
	
	return 0; 
}

void pause() {
	paused=1;
	mod.outMod->Pause(1);
}
void unpause() {
	RspAudioHLE(AudioHLE);
	paused=0;
	mod.outMod->Pause(0);
}
int ispaused() { return paused; }

void stop() {
	unpause();
	ShutdownApplication ();
}

// Only to be called within CPU thread
void InternalShutdown() {
	hCPU=0;
	paused=0;
	PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);

	ExitThread(0);
}

int getlength() {return TrackLength;}

int getoutputtime() {
	if (IsFreqSet()) return mod.outMod->GetOutputTime();
	else return 0;
}

void setoutputtime(int time_in_ms) {
	if (time_in_ms > getoutputtime()) SeekTo(time_in_ms);
	else DisplayError("Seeking backwards not supported");
}

void setvolume(int volume) {
	if (IsFreqSet()) mod.outMod->SetVolume(volume);
	SetVolume(volume);
}
void setpan(int pan) { if (IsFreqSet()) mod.outMod->SetPan(pan); }

BOOL CALLBACK rawDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_CLOSE:
		EndDialog(hDlg,TRUE);
		return 0;
	case WM_INITDIALOG:
		{
			char *LFpsftag;
			//char LFpsftag[500001];
			DWORD c,d;

			LFpsftag = malloc(strlen(infoDlgpsftag)*3+1);

			// <= includes terminating null
			for (c=0, d=0; c <= strlen(infoDlgpsftag); c++, d++) {
				LFpsftag[d]=infoDlgpsftag[c];
				if (infoDlgpsftag[c]=='\n') {
					LFpsftag[d++]='\r';
					LFpsftag[d]='\r';
					LFpsftag[d]='\n';
				}
			}
		
			SetDlgItemText(hDlg,IDC_RAWTAG,LFpsftag);
			//free(LFpsftag);
			break;
		}
	case WM_COMMAND:
		switch (GET_WM_COMMAND_ID(wParam, lParam)) {
		case IDOK:
			EndDialog(hDlg,TRUE);
			break;
		}
	default:
		return 0;
	}
	return 1;
}

BOOL CALLBACK infoDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	char Buffer[1024];
	char *LFpsftag;
	//char LFpsftag[50001];
	DWORD c,d;
	
	switch (uMsg) {	
	case WM_CLOSE:
		EndDialog(hDlg,TRUE);
		return 0;
	case WM_INITDIALOG:
		psftag_raw_getvar(infoDlgpsftag,"title",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_TITLE,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"artist",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_ARTIST,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"game",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_GAME,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"year",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_YEAR,Buffer);
		
		psftag_raw_getvar(infoDlgpsftag,"copyright",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_COPYRIGHT,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"usfby",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_USFBY,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"length",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_LENGTH,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"fade",Buffer,sizeof(Buffer)-1);
		SetDlgItemText(hDlg,IDC_FADE,Buffer);

		psftag_raw_getvar(infoDlgpsftag,"comment",Buffer,sizeof(Buffer)-1);

		LFpsftag = malloc(strlen(Buffer)*3+1);

		for (c=0, d=0; c <= strlen(Buffer); c++, d++) {
			LFpsftag[d]=Buffer[c];
			if (Buffer[c]=='\n') {
				LFpsftag[d++]='\r';
				LFpsftag[d]='\r';
				LFpsftag[d]='\n';
			}
		}

		SetDlgItemText(hDlg,IDC_COMMENT,LFpsftag);
		free(LFpsftag);

		break;
	case WM_COMMAND:
		switch (GET_WM_COMMAND_ID(wParam, lParam)) {
		case IDOK:
			GetDlgItemText(hDlg,IDC_TITLE,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"title",Buffer);
			
			GetDlgItemText(hDlg,IDC_ARTIST,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"artist",Buffer);
				
			GetDlgItemText(hDlg,IDC_GAME,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"game",Buffer);

			GetDlgItemText(hDlg,IDC_YEAR,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"year",Buffer);
			
			GetDlgItemText(hDlg,IDC_COPYRIGHT,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"copyright",Buffer);

			GetDlgItemText(hDlg,IDC_USFBY,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"usfby",Buffer);

			GetDlgItemText(hDlg,IDC_LENGTH,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"length",Buffer);

			GetDlgItemText(hDlg,IDC_FADE,Buffer,1024);
			psftag_raw_setvar(infoDlgpsftag,50000-1,"fade",Buffer);

			GetDlgItemText(hDlg,IDC_COMMENT,Buffer,1024);

			// remove 0x0d (in PSF a newline is 0x0a)
			for (c=0,d=0; c < strlen(Buffer); c++) {
				if (Buffer[c] != 0x0d) {
					Buffer[d]=Buffer[c];
					d++;
				}
			}
			Buffer[d]='\0';

			psftag_raw_setvar(infoDlgpsftag,50000,"comment",Buffer);

			infoDlgWriteNewTags=1;

			EndDialog(hDlg, TRUE);
			break;
		case IDC_CANCEL:
			EndDialog(hDlg, TRUE);
			break;
		case IDC_NOWBUT:
			sprintf(Buffer,"%i:%02i.%i",getoutputtime()/60000,(getoutputtime()%60000)/1000,getoutputtime()%1000);
			SetDlgItemText(hDlg,IDC_LENGTH,Buffer);
			break;
		case IDC_LAUNCHCONFIG:
			DialogBox(mod.hDllInstance, (const char *)IDD_CONFIG_WINDOW, hDlg, configDlgProc);
			break;
		case IDC_VIEWRAW:
			DialogBox(mod.hDllInstance, (const char *)IDD_RAWTAG_WINDOW, hDlg, rawDlgProc);
		}
		break;
	default:
		return 0;
	}

	return 1;
}

int infoDlg(char *fn, HWND hwnd)
{
	//DWORD c,d;
	//DWORD dwRead;
	//DWORD FileSize;
	//DWORD tagstart;
	//DWORD taglen;
	//DWORD reservesize;
	//HANDLE hFile;
	//BYTE Test[5];

	//hFile = CreateFile(fn,GENERIC_READ,FILE_SHARE_READ,NULL,
	//	OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
	//	NULL);

	//if (hFile == INVALID_HANDLE_VALUE) {
	//	return 1;
	//}

	//ReadFile(hFile,Test,4,&dwRead,NULL);
	//if (!IsValidUSF(Test)) {
	//	CloseHandle(hFile);
	//	return 1;
	//}
	// Read USF contents
	//ReadFile(hFile,&reservesize,4,&dwRead,NULL); // size of reserved area
	//ReadFile(hFile,&FileSize,4,&dwRead,NULL); // size of program
	//ReadFile(hFile,Test,4,&dwRead,NULL); // CRC (not checked yet)
	
	//tagstart=SetFilePointer(hFile,0,0,FILE_CURRENT)+FileSize+reservesize;
	//SetFilePointer(hFile,tagstart,0,FILE_BEGIN);

	//ReadFile(hFile,Test,5,&dwRead,NULL);
	// read tags if there are any
	//if (IsTagPresent(Test)) {
		//tagstart+=5;
		//taglen=SetFilePointer(hFile,0,0,FILE_END)-tagstart;
		
		infoDlgpsftag=malloc(50001); //malloc(taglen+2); // for safety when editing
		infoDlgpsftag[0]='\0';
		psftag_readfromfile(infoDlgpsftag,fn);
		
		//SetFilePointer(hFile,tagstart,0,FILE_BEGIN);
						
		//ReadFile(hFile,infoDlgpsftag,taglen,&dwRead,NULL);
		//CloseHandle(hFile);
		//infoDlgpsftag[taglen]='\0';

		infoDlgWriteNewTags=0;

		DialogBox(mod.hDllInstance, (const char *)IDD_INFO_WINDOW, hwnd, infoDlgProc);
		
		if (infoDlgWriteNewTags) {
			psftag_writetofile(infoDlgpsftag,fn);
		}
	//} else CloseHandle(hFile);
	free(infoDlgpsftag);
	infoDlgpsftag=NULL;

	return 0;
}

// Will work on any properly formatted string, will fail miserably on anything else
DWORD LengthFromString(const char * timestring) {
	int c=0,decimalused=0,multiplier=1;
	DWORD total=0;
	if (strlen(timestring) == 0) return 0;
	for (c=strlen(timestring)-1; c >= 0; c--) {
		if (timestring[c]=='.' || timestring[c]==',') {
			decimalused=1;
			total*=1000/multiplier;
			multiplier=1000;
		} else if (timestring[c]==':') multiplier=multiplier*6/10;
		else {
			total+=(timestring[c]-'0')*multiplier;
			multiplier*=10;
		}
	}
	if (!decimalused) total*=1000;
	return total;
}

void getfileinfo(char *filename, char *title, int *length_in_ms)
{
	DWORD dwRead;
	DWORD FileSize;
	DWORD tagstart;
	DWORD taglen;
	DWORD reservesize;
	HANDLE hFile;
	BYTE buffer[256],length[256],fade[256];
	BYTE Test[5];
	//char *psftag;
	
	if (!filename || !*filename) filename=lastfn;
	//if (title) sprintf(title,"%s",filename);


	hFile = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
		NULL);

	if (hFile == INVALID_HANDLE_VALUE) {
		return;
	}

	ReadFile(hFile,Test,4,&dwRead,NULL);
	if (!IsValidUSF(Test)) {
		CloseHandle(hFile);
		return;
	}
	// Read USF contents
	ReadFile(hFile,&reservesize,4,&dwRead,NULL); // size of reserved area
	ReadFile(hFile,&FileSize,4,&dwRead,NULL); // size of program
	ReadFile(hFile,Test,4,&dwRead,NULL); // CRC (no check yet)
	
	tagstart=SetFilePointer(hFile,0,0,FILE_CURRENT)+FileSize+reservesize;
	SetFilePointer(hFile,tagstart,0,FILE_BEGIN);

	ReadFile(hFile,Test,5,&dwRead,NULL);
	// read tags if there are any
	if (IsTagPresent(Test)) {
		tagstart+=5;
		taglen=SetFilePointer(hFile,0,0,FILE_END)-tagstart;
		SetFilePointer(hFile,tagstart,0,FILE_BEGIN);
		getfileinfopsftag=malloc(taglen+2);
				
		ReadFile(hFile,getfileinfopsftag,taglen,&dwRead,NULL);
		getfileinfopsftag[taglen]='\0';

		psftag_raw_getvar(getfileinfopsftag,"length",length,sizeof(length)-1);
		psftag_raw_getvar(getfileinfopsftag,"fade",fade,sizeof(fade)-1);

		if (length_in_ms) *length_in_ms=LengthFromString(length)+LengthFromString(fade);
		if (IgnoreTrackLength) *length_in_ms=0;
				
		// create title from PSF tag
		if (title) {
			DWORD c=0,tagnamestart;
			char tempchar;
			sprintf(title,"");
			while (c < strlen(titlefmt)) {
				if (titlefmt[c]=='%') {
					tagnamestart=++c;
					for (;c<strlen(titlefmt) && titlefmt[c] != '%'; c++);
					if (c == strlen(titlefmt)) {
						DisplayError("Bad title format string (unterminated token)");
						return;
					}
					titlefmt[c]='\0';
					psftag_raw_getvar(getfileinfopsftag,titlefmt+tagnamestart,buffer,sizeof(buffer)-1);
					strcat(title,buffer);
					titlefmt[c]='%';
				} else {
					tempchar=titlefmt[c+1];
					titlefmt[c+1]='\0';
					strcat(title,titlefmt+c);
					titlefmt[c+1]=tempchar;
				}
				c++;
			}
		}

		free(getfileinfopsftag);
		getfileinfopsftag=NULL;
	}
	if (DefaultLength && *length_in_ms <= 0) *length_in_ms=(deflen+deffade)*1000;
	CloseHandle(hFile);
}

void eq_set(int on, char data[10], int preamp) 
{ 
}

In_Module mod = 
{
	IN_VER,
	AppName,
	0,	// hMainWindow
	0,  // hDllInstance
	"usf\0Nintendo Ultra 64 Sound Format (*.usf)\0miniusf\0Mini-USF (*.miniusf)\0",
	1,	// is_seekable
	1, // uses output
	config,
	about,
	init,
	quit,
	getfileinfo,
	infoDlg,
	isourfile,
	play,
	pause,
	unpause,
	ispaused,
	stop,
	
	getlength,
	getoutputtime,
	setoutputtime,

	setvolume,
	setpan,

	0,0,0,0,0,0,0,0,0, // vis stuff


	0,0, // dsp

	eq_set,

	NULL,		// setinfo

	0 // out_mod

};

__declspec( dllexport ) In_Module * winampGetInModule2()
{
	return &mod;
}
