#include <windows.h>
#include <stdio.h>
#include <math.h>
#include "AudioCode.h"
#include "AudioSpec.h"

DWORD last_pos = 0, write_pos = 0, play_pos = 0, temp = 0;
DWORD last_write = -1;
//LPVOID lpvPtr1, lpvPtr2;
//DWORD dwBytes1, dwBytes2;
short sample_buffer[576*2*2]; // used for DSP
#define WINAMP_BUFFER_SIZE (576*4)
#define SILENTTHRESH 20
int freqset=0;
int volume;
DWORD silentcount=0;
DWORD buffsize = 0;
DWORD laststatus = 0;
double relvol = 1.0;
SHORT preamp[256*256];

void InitAmpTab(void) {
	unsigned long c;
	//DisplayError("InitAmpTab");

	if (softsaturate) {
		// Amplification is done via "saturation", out=1-(1-in)^(amp level). This approximates linear amp.
		// for low amplitudes but won't clip at high levels.

		for (c=0; c < 128*256; c++) {
			//preamp[(c^0xffff)+1]=-(preamp[c]=pow(1.0-fabs(pow(1.0-(((double)c)/(128.0*256.0)),relvol)),1.0/relvol)*(128.0*256.0));
			preamp[(c^0xffff)+1]=-(preamp[c]=(1.0-pow(1.0-((double)c)/(128.0*256.0),relvol))*(128.0*256.0));
		}
	} else {
		// clipping for those who really want it
		for (c=0; c < 128*256; c++) {
			preamp[(c^0xffff)+1]=-(preamp[c]=(relvol*c < 128*256)?(relvol*c):(128*256-1));
			//preamp[(c^0xffff)+1]=-(preamp[c]=relvol*c);
		}
	}
	// negative range is ever so slightly larger than positive range
	preamp[0x8000]=-128.0*256.0;
}

void AudioCode::InitWinampOutput(void) {
	int maxlatency;
	silentcount=0;
	if (!mod || !SampleRate) return;
	maxlatency=mod->outMod->Open(SampleRate,2,16, -1,-1);
	if (maxlatency < 0) {
		MessageBox(NULL,"Failed to open output plugin","Error",MB_OK);
		SampleRate=0;
		return;
	}
	mod->SetInfo(16,SampleRate/1000,2,1);
	mod->SAVSAInit(maxlatency,SampleRate);
	mod->VSASetInfo(SampleRate,2);
	//mod->SAVSAInit(maxlatency,44100);
	//mod->VSASetInfo(44100,2);
	mod->outMod->SetVolume(volume);
	freqset=1;
}
void AudioCode::CloseWinampOutput(void) {
	if (mod) {
		mod->outMod->Close();
		mod->SAVSADeInit();
		//MessageBox(NULL,"CloseWinampOutput","!",MB_OK);
	}
	freqset=0;
}

//------------------------------------------------------------------------

// Setup and Teardown Functions

// Generates nice alignment with N64 samples...
void AudioCode::SetSegmentSize (DWORD length) {
    if (SampleRate == 0) { return; }
	//WaitForSingleObject (hMutex, INFINITE);
	SegmentSize = length;
	//ReleaseMutex (hMutex);
}

void AiCallBack (DWORD n) {}

BOOL AudioCode::Initialize (void (*AiCallBack)( DWORD ), HWND hwnd) {
	return Initialize (hwnd);
}

BOOL AudioCode::Initialize (HWND hwnd) {
	audioIsPlaying = FALSE;
	freqset=0;
	silentcount=0;
	time=nexttime=0;
	DeInitialize (); // Release just in case...
	//hMutex = CreateMutex(NULL,FALSE,NULL);
	mod=AudioInfo.mod;
		
	return TRUE;
}

void AudioCode::DeInitialize () {
	//freqset=0;
	//if (hMutex) { 
	//	CloseHandle(hMutex); 
	//	hMutex = NULL;
	//}
	StopAudio ();
}

// ---------BLAH--------
#define u32 DWORD

//static u32 Frequency = 0;
//static u32 Length = 0;
//static double CountsPerByte;
//static u32 SecondBuff = 0;
//static u32 CurrentCount;
//static u32 CurrentLength;
//static u32 IntScheduled = 0;
//extern u32 VsyncTiming; // Need this so I can time things to VSync...


// Buffer Functions for the Audio Code
void AudioCode::SetFrequency (DWORD Frequency) {
	//SampleRate = (Frequency/100)*100;
	//DisplayError("SetFrequency");
	SampleRate = Frequency+25;
	SampleRate -= SampleRate % 50; // at Chris Moeller's suggestion
	SegmentSize = 0; // Trash it... we need to redo the Frequency anyway...
	SetSegmentSize (LOCK_SIZE);

	//freqset=1;
	//CloseWinampOutput();
	//InitWinampOutput();
}

DWORD lastword=0;

DWORD AudioCode::AddBuffer (BYTE *start, DWORD length) {
	DWORD retVal = 0,c;
	int t;
	static int seeklast=0;
	static short lastsample=0;

	//DisplayError("AddBuffer");
	
//	DWORD max = MAXBUFFER;
	if (length == retVal)
		return 0;

	if (!audioIsPlaying)
		StartAudio ();

	if (!playing())
		return retVal;
	
	time = nexttime;
	nexttime += (double)length*(double)1000.0/((double)SampleRate*4.0);
	
	if (nexttime >= seektime) {
		// Copy from AI buffer to internal buffer
		
		// process fade
		if (FadeLength && nexttime >= TrackLength-FadeLength) {
			for (c=0;c<length/2;c++) {
				if (time+c*2*1000.0/SampleRate < TrackLength-FadeLength) {
					// no fade yet
					//*((WORD*)(SoundBuffer+writeLoc))=((WORD*)start)[c];
					//*((short*)(SoundBuffer+writeLoc))=((double)((short*)start)[c^1])*relvol;
					*((short*)(SoundBuffer+writeLoc))=preamp[((unsigned short*)start)[c^1]];
				} else {
					// fade yet
					*((short*)(SoundBuffer+writeLoc))=
((double)preamp[((unsigned short*)start)[c^1]])*(TrackLength-(time+c*2*1000.0/SampleRate))/FadeLength;
				}
				if ((writeLoc+=2)==MAXBUFFER) writeLoc=0;
			}
		// insert silence after end of track
		} else if (TrackLength > 0 && nexttime >= TrackLength) {
			for (c=0;c<length/2;c++) {
				if (time+c*2*1000.0/SampleRate < TrackLength-FadeLength)
					// no fade, normal copy
					//*((WORD*)(SoundBuffer+writeLoc))=((WORD*)start)[c];
					*((short*)(SoundBuffer+writeLoc))=preamp[((unsigned short*)start)[c^1]];
				else
					// past end of track
					*((WORD*)(SoundBuffer+writeLoc))=0;
				if ((writeLoc+=2)==MAXBUFFER) writeLoc=0;
			}
		// no end of track condition
		} else {
			if (!DetectSilence) {
				for (c=0;c<length/2;c++) {
					//*((WORD*)(SoundBuffer+writeLoc))=((WORD*)start)[c];
					*((short*)(SoundBuffer+writeLoc))=preamp[((unsigned short*)start)[c^1]];
					if ((writeLoc+=2)==MAXBUFFER) writeLoc=0;
				}
			} else {
				for (c=0;c<length/2;c++) {
					/**((short*)(SoundBuffer+writeLoc))=((double)((short*)start)[c^1]*relvol);
					if ((writeLoc+=2)==MAXBUFFER) writeLoc=0;
					if (((short*)start)[c]>-SILENTTHRESH && ((short*)start)[c]<SILENTTHRESH) silentcount++;
					else silentcount=0;
					if (silentcount >= silencelength*SampleRate*2) silencedetected=1;*/
					
					if (abs(lastsample-((short*)start)[c^1]) < SILENTTHRESH) {
						silentcount++;
					} else {
						silentcount=0;
						// only update last sample when no silence, this should prevent a continuous
						// small delta from being detected as silence
						lastsample=*((short*)(SoundBuffer+writeLoc))=((short*)start)[c^1];
					}
					*((short*)(SoundBuffer+writeLoc))=preamp[((unsigned short*)start)[c^1]];
					if ((writeLoc+=2)==MAXBUFFER) writeLoc=0;
										
					if (silentcount >= silencelength*SampleRate*2) silencedetected=1;
				}
			}
		}

		remainingBytes+=length;
				
		// Send from internal buffer to Winamp
		while (mod->outMod->CanWrite() < (remainingBytes/WINAMP_BUFFER_SIZE)*WINAMP_BUFFER_SIZE)
			Sleep(50);
		while (remainingBytes >= WINAMP_BUFFER_SIZE) {
			t=mod->outMod->GetWrittenTime();
			mod->SAAddPCMData((char*)SoundBuffer+readLoc,2,16,t);
			mod->VSAAddPCMData((char*)SoundBuffer+readLoc,2,16,t);
						
			// DSP?
			if (mod->dsp_isactive()) {
				int samples;
				memcpy(sample_buffer,(char*)SoundBuffer+readLoc,WINAMP_BUFFER_SIZE);
				samples=mod->dsp_dosamples(sample_buffer,576,16,2,SampleRate);
				mod->outMod->Write((char*)sample_buffer,samples*4);
			} else mod->outMod->Write((char*)SoundBuffer+readLoc,WINAMP_BUFFER_SIZE);
			remainingBytes-=WINAMP_BUFFER_SIZE;
			if ((readLoc+=WINAMP_BUFFER_SIZE)==MAXBUFFER) readLoc=0;
		}
		if (seeklast) {
			//SetThreadPriority(hCPU, THREAD_PRIORITY_HIGHEST);
			SetThreadPriority(hCPU, priarray[CPUPriority]);
		}
		seeklast=0;
	} else {
		mod->outMod->Flush((int)time);
		writeLoc = 0x0000;
		readLoc = 0x0000;
		remainingBytes = 0;
		if (!seeklast) SetThreadPriority(hCPU, THREAD_PRIORITY_NORMAL);
		seeklast=1;
	}
	
	//if (time >= (10*1000)) {
		//StopAudio();
		//PostMessage(mod->hMainWindow,WM_WA_MPEG_EOF,0,0);
		//CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)(mod->Stop),0,0,&ThreadID);
		//return 1;
	//}

	return retVal;
}

// Management functions
// TODO: For silent emulation... the Audio should still be "processed" somehow...
void AudioCode::StopAudio () {
	//DisplayError("StopAudio");
	if (!audioIsPlaying) return;
	audioIsPlaying = FALSE;
	//TerminateThread (this->handleAudioThread, 0);
	CloseWinampOutput();
}

void AudioCode::StartAudio () {
	if (audioIsPlaying) return;
	audioIsPlaying = TRUE;	

	InitWinampOutput();

	//this->handleAudioThread = CreateThread (NULL, NULL, (LPTHREAD_START_ROUTINE)AudioThreadProc, this, NULL, &this->dwAudioThreadId);
	writeLoc = 0x0000;
	readLoc = 0x0000;
	remainingBytes = 0;
	seektime=time=nexttime=0;

	InitAmpTab();
}

EXPORT DWORD CALL IsFreqSet(void) {
	return freqset;
}

DWORD AudioCode::GetReadStatus () {
	return 0;
}

EXPORT void CALL SetVolume(int newvol) {
	volume=newvol;
}
